/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Gmail Notifier code.
 *
 * The Initial Developer of the Original Code is
 * Doron Rosenberg.
 * Portions created by the Initial Developer are Copyright (C) 2004 - 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const kGMSERVICE_CONTRACTID = "@mozilla.org/GMailNotifier;1"
const kGMSERVICE_CID = Components.ID("02509a0c-249e-4396-8f72-fdf5af33d7b4");
const nsINotifierService = Components.interfaces.nsIGMNotifierService
const nsINotifierProgressListener = Components.interfaces.nsIGMNotifierProgressListener;
const nsISupports = Components.interfaces.nsISupports;

const kIOSERVICE_CONTRACTID = "@mozilla.org/network/io-service;1";
const nsIIOService = Components.interfaces.nsIIOService;

const nsICookieManager = Components.interfaces.nsICookieManager;
const kCOOKIESERVICE_CONTRACTID = "@mozilla.org/cookieService;1";

const kPSWDMANAGER_CONTRACTID = "@mozilla.org/passwordmanager;1";
const nsIPasswordManagerInternal = Components.interfaces.nsIPasswordManagerInternal;

const kTIMER_CONTRACTID = "@mozilla.org/timer;1";
const nsITimer = Components.interfaces.nsITimer;

function nsNotifierService() {
  // log string
  this.logString = "";
  // array of attached listeners
  this.listeners = new Array();
  this.username = null;
  this.password = null;
  this.notificationListenerID = null;

  this.resetState = false;
  this.connectionPhase = null;
  this.isAutoConnecting = false;
  this.connectionCounter = 0;

  this.prefBranch = null;

  this.folders = new Array();

  this.unreadEmails = null;
  this.tmpUnreadEmails = null;

  this.newEmails = 0;
  this.space_used_mb = null;
  this.space_used_percent = null;
  this.total_mb = null;
  this.inboxUnread = 0;
  this.inboxNew = 0;

  this.updateTimer = null;
  this.timeOut = 600000; // default to 10 mins

  var myThis = this;
  this.PrefChangeObserver = {
    observe: function(aSubject, aTopic, aData)
    {
      myThis.prefChanged(aData);
    }
  };

  this.addObserver("gm-notifier", this.PrefChangeObserver);
}

/********** Scriptable interfaces ****************/

/**
 * Initiates login
 *
 * @param aUsername
 * @param aPassword
 * @param aListenerID
 */
nsNotifierService.prototype.initLogin = function(aUsername, aPassword, aListenerID) {
  gNotifierService.logItem("initLogin: aListenerID is " + aListenerID);
  gNotifierService.username = aUsername;
  gNotifierService.password = aPassword;

  gNotifierService.notificationListenerID = aListenerID;

  gNotifierService.startConnection();
}

/**
 * Initiates a new mail check
 *
 */
nsNotifierService.prototype.checkNow = function() {
  gNotifierService.startConnection();
}

/**
 * Logs the user out
 *
 */
nsNotifierService.prototype.logout = function() {
  this.username = null;
  this.password = null;

  this.unreadEmails = null;
  this.newEmails = 0;
  this.space_used_mb = null;
  this.space_used_percent = null;
  this.total_mb = null;

  // clear the timer
  if (this.updateTimer)
    this.updateTimer.cancel();
  this.updateTimer = null;

  this.loggedIn = false;
  this.connectionCounter = 0;
  this.remembermeCheck = false;

  // clear folders
  this.folders = new Array();
  this.inboxUnread = 0;

  // clear the login cookie
  try {
    var ios = Components.classes[kIOSERVICE_CONTRACTID].getService(nsIIOService);
    var cookieMgr = Components.classes[kCOOKIESERVICE_CONTRACTID].getService(nsICookieManager);
    cookieMgr.remove("mail.google.com", "GX", "/mail", false); cookieMgr.remove(".google.com", "SID", "/", false);
  } catch(e) {}

  // tell all listeners that we logged out
  this.pushStateChange(nsINotifierProgressListener.LOGOUT);
}

/**
 * Adds an NotifierListener to this service
 *
 * @param aListener
 */
nsNotifierService.prototype.addListener = function(aListener) {
  // should we autologin?
  // make sure to only check for autologin the first time a listener has been
  // added, ie first browser window calls us.
  if ((this.listeners.length == 0) && this.getPrefBranch() &&
      this.getPrefBranch().getBoolPref("gm-notifier.autologin.enabled")) {
    // autologin
    var defaultUser = this.getPrefBranch().getCharPref("gm-notifier.users.default");
    if (defaultUser) {
      this.username = defaultUser;
      this.password = this.getLoginDetails(this.username);
      this.isAutoConnecting = true;
      this.startConnection();
    }
  }

  this.listeners.push(aListener);

  // for every new listener added, push login state
  if (this.loggedIn)
    aListener.onStateChange(nsINotifierProgressListener.LOGIN_SUCCESS);
  else
    aListener.onStateChange(nsINotifierProgressListener.LOGOUT);

  return this.listeners.length;
}

/**
 * Removes an NotifierListener to this service
 *
 * @param aListener
 */
nsNotifierService.prototype.removeListener = function(aListener) {
  var found = false;
  var run = 0;

  while (!found || (run < this.listeners.length)) {
    if (this.listeners[run] == aListener) {
      this.listeners[run] = null;
      found = true;
    }
    run++;
  }
}

/**
 * Gets the reset state.  Reset is set if the user clicks/checks for new mail.
 */
nsNotifierService.prototype.getResetState = function() {
  return this.resetState;
}

/**
 * Sets the reset state
 *
 * @param aResetState
 */
nsNotifierService.prototype.setResetState = function(aResetState) {
  this.resetState = aResetState;
  gNotifierService.pushStateChange(nsINotifierProgressListener.LOGIN_SUCCESS);
}

nsNotifierService.prototype.getInboxUnread = function() {
  return this.inboxUnread;
}

nsNotifierService.prototype.getUnreadCount = function() {
  return this.unreadEmails;
}

nsNotifierService.prototype.getDisplayCount = function() {
  if (this.resetState) {
    return 0;
  } else {
    if (this.prefBranch.getBoolPref("gm-notifier.ui.counter.showInbox"))
     return this.getInboxUnread();
    else
      return this.unreadEmails;
  }
}

nsNotifierService.prototype.getNewCount = function() {
  var value;

  if (this.resetState) {
    value = 0;
  } else {
    if (this.prefBranch.getBoolPref("gm-notifier.ui.counter.showInbox"))
      value = this.inboxNew;
    else
      value = this.newEmails;
  }

  return value;
}

nsNotifierService.prototype.getUsedMB = function() {
  return this.space_used_mb;
}

nsNotifierService.prototype.getSpaceUsed = function() {
  return this.space_used_percent;
}

nsNotifierService.prototype.getTotalSpace = function() {
  return this.total_mb;
}

nsNotifierService.prototype.getUserName = function() {
  return this.username;
}

nsNotifierService.prototype.getFolderCount = function() {
  return this.folders.length;
}

nsNotifierService.prototype.getFolderItem = function(aPosition, sizeObj) {
  if (!this.folders[aPosition])
    return null;

  sizeObj.value = 2;
  return [this.folders[aPosition].name, this.folders[aPosition].unreadMail];
}

nsNotifierService.prototype.setTimeout = function(aMinutes) {
  // convert to ms
  this.timeOut = (aMinutes * 60000);

  // set timer only if we are logged in
  if (this.loggedIn)
    this.setTimer();
}

/**
 * Adds an item to the log
 *
 */
nsNotifierService.prototype.logItem = function(aLog) {
  this.logString += aLog + "|||";
  // dump("\n::::: " + aLog + "\n");
}

nsNotifierService.prototype.getLog = function() {
  return this.logString;
}

/*************** Private Methods ************/


nsNotifierService.prototype.talkToServer = function(aURL, aPostData,
                                                    aCookieData, aReferrer,
                                                    aCallbackFunc) {
  gNotifierService.logItem("  talkeToServer Called");

  // the IO service
  var ioService = Components.classes[kIOSERVICE_CONTRACTID].getService(nsIIOService);

  // create an nsIURI
  var uri = ioService.newURI(aURL, null, null);

  //nsIInputStream
  var uploadStream = Components.classes["@mozilla.org/io/string-input-stream;1"]
    .createInstance(Components.interfaces.nsIStringInputStream);

  if (aPostData)
    uploadStream.setData(aPostData, aPostData.length);

  // get a channel for that nsIURI
  var channel = ioService.newChannelFromURI(uri);

  // get a httpchannel and make it a post
  var httpChannel = channel.QueryInterface(Components.interfaces.nsIHttpChannel);

  // set a referrer
  if (aReferrer) {
    var referrerUri = ioService.newURI(aReferrer, null, null);
    httpChannel.referrer = referrerUri;
  }

  if (aPostData) {
    var uploadChannel = channel.QueryInterface(Components.interfaces.nsIUploadChannel);
    uploadChannel.setUploadStream(uploadStream, "application/x-www-form-urlencoded", -1);

    // order important - setUploadStream resets to get/put
    httpChannel.requestMethod = "POST";
  }

  if (aCookieData){
    // httpChannel.setRequestHeader("Cookie", aCookieData, false);
    for (var run = 0; run < aCookieData.length; run++){
      httpChannel.setRequestHeader("Cookie", aCookieData[run], true);
    }
  }

  var observer = new this.observer(aCallbackFunc);
  channel.asyncOpen(observer, null);
}


nsNotifierService.prototype.accountCheckComplete = function(aStatus) {
  /* let accountCheckComplete determine if we are supposed to send logout or not
     if invalid username/password logout
     if wrong page arrived, don't do anything and continue with timer.
       - but if initial login, still fail
        - if auto-login, keep trying
   */

  if (aStatus == nsINotifierProgressListener.LOGIN_FAILED) {
    gNotifierService.pushStateChange(aStatus);
  } else if (aStatus == nsINotifierProgressListener.LOGIN_DETAILS_INVALID) {
    gNotifierService.logItem("  Login details were invalid");
    gNotifierService.pushStateChange(nsINotifierProgressListener.LOGIN_FAILED);
  }

  // no longer auto connecting
  gNotifierService.isAutoConnecting = false;
  gNotifierService.connectionCounter++;
}

/**
 * startConnection and callback are the Gmail specific parts.
 */ 
nsNotifierService.prototype.startConnection = function () {

  gNotifierService.logItem("  StartConnection: Init");

  if (!this.username || !this.password) {
    return;
  }

  // clear the log every 2 checks
  if (this.connectionCounter == 2) {
    this.connectionCounter = 0;
    this.logString = "";
  }

  gNotifierService.logItem("  -- Username/Password found");

  this.resetState = false;

  this.connectionPhase = 0;
  this.pushStateChange(nsINotifierProgressListener.LOGIN_INITIATED);

  // has to be before the if check, since the cookie clearing code needs it
  // below
  var cookieMgr = Components.classes[kCOOKIESERVICE_CONTRACTID]
                            .getService(Components.interfaces.nsICookieManager);

  // find the google.com rememberme cookie and reuse it for the login data
  // do this check only the first time.
  if (!gNotifierService.remembermeCheck) {
    gNotifierService.remembermeCheck = true;
    gNotifierService.rememberme = "";

    var e = cookieMgr.enumerator;
    var cookie, done = false;
    while (!done && e.hasMoreElements()) {
      cookie = e.getNext().QueryInterface(Components.interfaces.nsICookie);

      if (cookie.host == ".google.com") {
        if (cookie.name == "rememberme") {
          done = true;
          gNotifierService.rememberme = (cookie.value == "true" ? "&PersistentCookie=true" : "");
          gNotifierService.logItem("  -- rememberme cookie value is " + cookie.value);
        } else if (cookie.name == "LSID" && (gNotifierService.rememberme == "")) {
          // no remember me, check expire date
          var expireDate = new Date(1000 * cookie.expires);
          var currentDate = new Date();

          gNotifierService.logItem("  "+expireDate.getFullYear() + " vs "  + currentDate.getFullYear());

          if ((expireDate.getFullYear() - 2) > currentDate.getFullYear()) {
            // expires in more than 2 years
            gNotifierService.rememberme = "&PersistentCookie=true";
            gNotifierService.logItem("  -- rememberme cookie value is " + cookie.value);
          }
        }
      }
    }

    // if rememberme is set, we should clear the login cookies if a new user
    // is being logged in.
    if (gNotifierService.rememberme) {
      var defaultUser = this.getPrefBranch().getCharPref("gm-notifier.users.default");
      if (defaultUser != this.username) {
        cookieMgr.remove("mail.google.com", "GX", "/mail", false);
      }
    }
  }

  var urlprotocol = this.getURLProtocol();

  // is it a hosted domain account?  Make sure the domain isn't gmail.com or
  // googlemail.com

  if (this.username.indexOf("@") > 0) {
    var index = this.username.indexOf("@");
    var domain = this.username.substring(index+1, this.username.length);
    if (domain == "gmail.com" || domain == "googlemail.com")
      this.username = this.username.substring(0, index);
  }

  if (this.username.indexOf("@") > 0) {
    var index = this.username.indexOf("@");
    var realusername = this.username.substring(0, index);
    this.hosteddomain = this.username.substring(index+1, this.username.length);
    gNotifierService.logItem("  -- hosted domain found: " + this.hosteddomain);

    var data = "at=null&continue=" + urlprotocol + "mail.google.com/hosted/"+this.hosteddomain+"/" + "&service=mail&userName=" + encodeURIComponent(realusername)
               + "&password=" + encodeURIComponent(this.password) + "&persistent=true";

    this.talkToServer("https://www.google.com/a/"+this.hosteddomain+"/LoginAction", data,
                      null, "https://www.google.com/a/"+this.hosteddomain+"/ServiceLogin", this.callback);
  } else {
    this.hosteddomain = null;
    var data = "?service=mail&Email=" + encodeURIComponent(this.username)
             + "&Passwd=" + encodeURIComponent(this.password) +
             gNotifierService.rememberme + "&rm=false&null=Sign%20in&continue=" + urlprotocol + "mail.google.com/mail?ui=html&amp;zy=l";
    this.talkToServer("https://www.google.com/accounts/ServiceLoginAuth", data,
                      null, "https://www.google.com/accounts/ServiceLogin", this.callback);
  }
}

/*
   Phase              Description
   0                  Before anything has happend
   1                  First connection
   ...
*/

nsNotifierService.prototype.callback = function(aData, aRequest) {
  gNotifierService.connectionPhase++;

  gNotifierService.logItem("    callback called, phase" + gNotifierService.connectionPhase);

  var urlprotocol = gNotifierService.getURLProtocol();

  switch (gNotifierService.connectionPhase) {
    case 1:
      var cookieData;

      // if no cookies sent, an exception is thrown
      /* try{
          var httpChannel = aRequest.QueryInterface(Components.interfaces.nsIHttpChannel)
          cookieData = httpChannel.getRequestHeader("cookie");
        } catch(e){} */
      var val = aData.indexOf("location.replace");

      if (val < 0) {
        // XXX - check if its an image request

        // check if it is a password re-request
        if (aData.indexOf("<form action=\"LoginAuth\"") >= 0)
          gNotifierService.accountCheckComplete(nsINotifierProgressListener.LOGIN_DETAILS_INVALID);
        else
          gNotifierService.accountCheckComplete(nsINotifierProgressListener.LOGIN_FAILED);

        // failed to login
        //gNotifierService.pushStateChange(nsINotifierProgressListener.LOGIN_FAILED);
        return;
      }

      //var url =  val[0].substr(2, (val[0].length-3) );
      /* var val = aData.match(/var cookieVal= "([^"]+)";/); */

      var url = urlprotocol;
      if (gNotifierService.hosteddomain)
        url +="mail.google.com/hosted/" + gNotifierService.hosteddomain;
      else
        url += "mail.google.com/mail";

      gNotifierService.talkToServer(url + "?ui", null, null, url+"/",
                                    gNotifierService.callback);
      break;

    case 2:

      var url = urlprotocol + encodeURIComponent(this.username) + ":" + encodeURIComponent(this.password) + "@mail.google.com/";
      if (gNotifierService.hosteddomain)
        url +="hosted/" + gNotifierService.hosteddomain;
      else
        url += "mail";

      gNotifierService.talkToServer(url + "/feed/atom/unread",
                                    null, null, urlprotocol + "mail.google.com/",
                                    gNotifierService.callback);
      break;

    case 3:
      gNotifierService.logItem("  connecting to gmail rss feed");

      // the feed was loaded
      var val = aData.match(/<fullcount>(.*)<\/fullcount>/);

      /*var tmp = aData;
      tmp = tmp.replace(/</g, "&lt;");
      gNotifierService.logItem("Feed is:" + tmp);*/

      if (val) {
        gNotifierService.logItem("  -  Atom feed for unread returned " + val[1] + " unread.");
        gNotifierService.tmpUnreadEmails = val[1];
      } else {
        gNotifierService.logItem("  -  Getting the Atom feed failed!");
      }

      var url = urlprotocol;
      if (gNotifierService.hosteddomain)
        url +="mail.google.com/hosted/" + gNotifierService.hosteddomain;
      else
        url += "mail.google.com/mail";

      gNotifierService.talkToServer(url+ "/?search=inbox&view=tl&start=0&init=1",
                                    null, null, url, gNotifierService.callback);
      break;

    case 4:
      gNotifierService.logItem("  connecting to gmail complete.");

      // Gmail now serves 2 versions of the ds string:
      // D(["ds",46,0,0,0,0,1298,0] or
      // D(["ds",[["inbox",3]\n,["drafts",2]\n,["spam",1053]\n]\n]n);

      var val;
      var newDS = false;

      if (aData.match(/\["ds",\[\[/)) {
        // new way
        val = aData.match(/\["ds",.*\n(.*]\n)+\]/gm);

        newDS = true;
      } else {
        val = aData.match(/\["ds",.*\]/);
      }

      if (!val) {
        // failed to login
        gNotifierService.logItem("    - Login failed, final page not the correct one.");
        gNotifierService.pushStateChange(nsINotifierProgressListener.LOGIN_FAILED);
        return;
      }

      // get the unread count
      var ts = aData.match(/\["ts",.*\]/gm);
      var myArray = eval(ts[0]);
      gNotifierService.logItem("  -  ts is " + ts);

      // utf-8 fun
      var conv = Components.classes["@mozilla.org/intl/utf8converterservice;1"].getService(Components.interfaces.nsIUTF8ConverterService);

      // get the inbox label (localized) from the website
      var inboxLabel = escape(conv.convertStringToUTF8(myArray[5], "utf-8", false));

      var totalUnread;
      // XXX: Gmail lies to us, so don't trust this for now
      // however, use it if the tmpUnreadEmails was null, ergo we failed to load
      // the atom feed.
      if (gNotifierService.tmpUnreadEmails != null) {
        totalUnread = gNotifierService.tmpUnreadEmails;
      } else {
        totalUnread = myArray[3];
        gNotifierService.logItem("  -  myArray[3] " +  myArray[3]);
      }

      // handle inbox
      myArray = eval(val[0]);
      gNotifierService.logItem("  -  val[0] " +  val[0]);

      var inboxUnread = 0;
      if (newDS) {
        inboxUnread = myArray[1][0][1];
      } else {
        inboxUnread = myArray[1];
      }

      gNotifierService.logItem("  -  inboxUnread " +  inboxUnread + " and gNotifierService.inboxUnread is " + gNotifierService.inboxUnread);
      gNotifierService.inboxNew = inboxUnread - gNotifierService.inboxUnread;
      gNotifierService.inboxUnread = inboxUnread;

      gNotifierService.folders = new Array();
      gNotifierService.folders[0] = gNotifierService.createFolder(inboxLabel, inboxUnread);

      if (gNotifierService.inboxNew < 0)
        gNotifierService.inboxNew = 0;

      // handle labels
      var labelsVal = aData.match(/\["ct",.*\n(.*]\n)+\]/gm);
      if (labelsVal) {
        var labelsArray = eval(labelsVal[0])[1];

        for (var run = 0; run < labelsArray.length; run++) {
          var folderName = escape(conv.convertStringToUTF8(labelsArray[run][0], "utf-8", false));

          // gmail skins extension workaround
          var skip = (folderName.indexOf("gmskin%3A") == 0);

          if (!skip) {
            gNotifierService.folders.push(gNotifierService.createFolder(folderName, labelsArray[run][1]));
            gNotifierService.logItem("  - Folder found (name: " + gNotifierService.folders[gNotifierService.folders.length-1].name + " and unread emails: " +  gNotifierService.folders[[gNotifierService.folders.length-1]].unreadMail + ")");
          }
        }
      }

      // new/unread mail check
      gNotifierService.logItem("  - GMail new mail check complete");
      gNotifierService.logItem("  -inboxUnread is " + inboxUnread + " unread is " + gNotifierService.unreadEmails);

      var showInboxOnly = gNotifierService.prefBranch.getBoolPref("gm-notifier.ui.counter.showInbox");
      var hasNewMail = (totalUnread > gNotifierService.unreadEmails);

      if (!hasNewMail && showInboxOnly) {
        hasNewMail = (gNotifierService.inboxUnread > 0);
      }

      if ((gNotifierService.unreadEmails != null) && hasNewMail) {
        gNotifierService.logItem("  New Unread Mail Found!");
        // new email
        gNotifierService.newEmails = totalUnread - gNotifierService.unreadEmails;

        if (showInboxOnly && (gNotifierService.inboxUnread == 0)) {
          // show inbox only, but new mail wasn't in inbox
          gNotifierService.pushStateChange(nsINotifierProgressListener.N0_NEW_MAIL);
        } else {
          gNotifierService.pushStateChange(nsINotifierProgressListener.NEW_MAIL);
          gNotifierService.newMailNotification(gNotifierService.getNewCount());
        }
      } else if (gNotifierService.unreadEmails == null) {
        // first time
        if (totalUnread > 0) {
          gNotifierService.logItem("  New Mail Found!");
          gNotifierService.newEmails = totalUnread;
          gNotifierService.unreadEmails = totalUnread;
          gNotifierService.pushStateChange(nsINotifierProgressListener.NEW_MAIL);
          gNotifierService.logItem("  - amount of new mail is " + gNotifierService.newEmails);
          gNotifierService.logItem("  - mail notification about to be set, new mail count is " +  gNotifierService.getDisplayCount());
          gNotifierService.newMailNotification(gNotifierService.getDisplayCount());
        } else {
          gNotifierService.newEmails = 0;
          gNotifierService.unreadEmails = 0;
        }
      } else {
        gNotifierService.newEmails = 0;
        gNotifierService.logItem("  No New Unread Mail Found!");
        gNotifierService.pushStateChange(nsINotifierProgressListener.NO_NEW_MAIL);
      }

      gNotifierService.unreadEmails = totalUnread;

      val = aData.match(/\["qu",.*\]/);
      myArray = eval(val[0]);

      // localized versions of gmail add % and MB for some reason!
      gNotifierService.space_used_mb = myArray[1];
      var length = myArray[1].length;
      if (myArray[1][length-3] == " " && myArray[1][length-2] == "M" && myArray[1][length-1] == "B")
        gNotifierService.space_used_mb = myArray[1].substr(0,length-3);
      else
        gNotifierService.space_used_mb = myArray[1];

      length = myArray[2].length;
      if (myArray[2][length-3] == " " && myArray[2][length-2] == "M" && myArray[2][length-1] == "B")
        gNotifierService.total_mb = myArray[2].substr(0,length-3);
      else
        gNotifierService.total_mb = myArray[2];

      length = myArray[3].length;
      if (myArray[3][length-1] == "%")
        gNotifierService.space_used_percent = myArray[3].substr(0,length-1);
      else
        gNotifierService.space_used_percent = myArray[3];

      gNotifierService.loggedIn = true;
      gNotifierService.connectionCounter++;

      gNotifierService.pushStateChange(nsINotifierProgressListener.LOGIN_SUCCESS);
      //gNotifierService.setTimer();
      break;
  }
}


// timing stuff
nsNotifierService.prototype.setTimer = function() {
  if (this.updateTimer) {
    this.updateTimer.cancel();
  } else {
    this.updateTimer = Components.classes[kTIMER_CONTRACTID].createInstance(nsITimer);
  }

  gNotifierService.logItem("Starting at : " + new Date());
  try{
    gNotifierService.updateTimer.initWithCallback(gNotifierService,
                                                    gNotifierService.timeOut,
                                                    nsITimer.TYPE_ONE_SHOT);
  } catch (e) {
    gNotifierService.logItem(e);
  }
}

// nsITimer
nsNotifierService.prototype.notify = function(aTimer) {
  gNotifierService.logItem("  *************** notify called");
  gNotifierService.logItem("End: " + new Date());
  gNotifierService.startConnection();
}

nsNotifierService.prototype.pushStateChange = function(aState) {
  for (var run = 0; run < this.listeners.length; run++)
    if (this.listeners[run]) // can be null if a listener was removed
      this.listeners[run].onStateChange(aState);
}

nsNotifierService.prototype.observer = function(aCallbackFunc) {
  return ({
    data : "",

    onStartRequest : function (aRequest, aContext) {
      this.data = "";
    },

    onDataAvailable : function (aRequest, aContext, aStream, aSourceOffset, aLength){
      var scriptableInputStream = 
        Components.classes["@mozilla.org/scriptableinputstream;1"]
          .createInstance(Components.interfaces.nsIScriptableInputStream);
      scriptableInputStream.init(aStream);

      this.data += scriptableInputStream.read(aLength);
    },

    onStopRequest : function (aRequest, aContext, aStatus) {
      aCallbackFunc(this.data, aRequest);
    },

    QueryInterface : function(aIID) {
      if (aIID.equals(Components.interfaces.nsIStreamListener) ||
          aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
          aIID.equals(Components.interfaces.nsIAlertListener) ||
          aIID.equals(nsISupports))
        return this;
      throw Components.results.NS_NOINTERFACE;
    }
  }
  );
}

/*
  New mail notification
 */
nsNotifierService.prototype.newMailNotification = function(aNewNum) {
  if (aNewNum < 1)
    return;

  this.logItem("New Mail Notification Init:");

  var isNotificationEnabled = this.getPrefBranch().getBoolPref("gm-notifier.ui.notification.enabled");
  this.logItem("  Is System Notification Enabled by user: " + isNotificationEnabled);

  if (isNotificationEnabled){
    var msg = this.getFormattedString("NotificationMsg", [aNewNum]);

    try {
      var alertService = Components.classes["@mozilla.org/alerts-service;1"]
                                   .getService(Components.interfaces.nsIAlertsService);
      if (alertService) {
        alertService.showAlertNotification("chrome://gm-notifier/content/gm-logo.png",
          this.getString("NotificationMsgTitle"), msg, true, "", this);
        this.logItem("  alertsService success.");
      } else {
        this.logItem("  alertsService failure: could not getService nsIAlertsService");
      }
    } catch(e) {
       this.logItem("  alertsService failure: " + e);
    }
  }

  // sound notifications
  var isSoundNotificationsEnabled = this.getPrefBranch().getBoolPref("gm-notifier.ui.soundnotification.enabled");
  this.logItem("  Is Sound Notification Enabled by user: " + isSoundNotificationsEnabled);

  if (isSoundNotificationsEnabled) {
    var soundUrl = this.getPrefBranch().getCharPref("gm-notifier.ui.soundnotification.uri");

    try {
      var sound = Components.classes["@mozilla.org/sound;1"]
                            .createInstance(Components.interfaces.nsISound);
      // handle locale files using the miracle of XPCOM
      if (soundUrl.indexOf("file://") == -1) {
        var filePath = new Components.Constructor("@mozilla.org/file/local;1",
                                                  "nsILocalFile",
                                                  "initWithPath");
        var file = new filePath(soundUrl);

        var ioService = Components.classes['@mozilla.org/network/io-service;1']
                                  .getService(Components.interfaces.nsIIOService);
        url = ioService.newFileURI(file, null, null);
      } else {
        var url = Components.classes["@mozilla.org/network/standard-url;1"]
                            .createInstance(Components.interfaces.nsIURL);
        url.spec = soundUrl;
      }

      sound.play(url)
    } catch (e) {
      this.logItem("  Sound playing failed with: "+ e);
    }
  }
}

// nsIObserver for alert service
nsNotifierService.prototype.observe = function (aSubject, aTopic, aData){
  switch (aTopic) {
    case "alertfinished":
      break;
                   
    case "alertclickcallback":
      this.showNotification();
      break;
  }
}

// nsIAlertsService - nsIAlertListener was deprecated and replaced with nsIObserver
//                    This is to maintain 1.7 compatability.
nsNotifierService.prototype.onAlertFinished = function (aCookie){
}

nsNotifierService.prototype.onAlertClickCallback = function (aCookie){
  this.showNotification();
}

nsNotifierService.prototype.showNotification = function (){
  // when the alert callback happens, we want to notify the listener that
  // initiated the server connection that the alert was clicked.
  // Simply pushing a state change to all listeners will result in the
  // infinite window open loop!
  
  this.logItem("Alert Clickback called:");
  this.logItem("  Notification Window is: (" + gNotifierService.notificationListenerID + ").");

  var run = 0;
  if (gNotifierService.notificationListenerID) {
    var done = false;
    while (!done && (run < this.listeners.length)) {
      this.logItem("  Item " + run +  " has id (" + this.listeners[run].getID() + ")");
      if (this.listeners[run] && (this.listeners[run].getID() == this.notificationListenerID))
        done = true;
      else 
        run++;
    }
  } else {
    // fallback
    while ((this.listeners[run] == null) && (run < this.listeners.length))
      run++;

    this.logItem("  No notification window found, call listener #" + run);
  }

  if (this.listeners[run])
    this.listeners[run].onStateChange(nsINotifierProgressListener.LOAD_MAIL);
}

nsNotifierService.prototype.createFolder = function(aName, aUnreadMail) {
  return {
    name: aName,
    unreadMail: aUnreadMail
  }
}

/* Pref code*/
nsNotifierService.prototype.getPrefBranch = function(){
  if (!this.prefBranch){ 
    this.prefBranch = Components.classes['@mozilla.org/preferences-service;1'];
    this.prefBranch = this.prefBranch.getService();
    this.prefBranch = this.prefBranch.QueryInterface(Components.interfaces.nsIPrefBranch);
  }
  
  return this.prefBranch;
}

nsNotifierService.prototype.prefChanged = function(aPrefName) {
  switch(aPrefName){
    case "gm-notifier.update.interval":
      this.setTimeout(this.prefBranch.getIntPref("gm-notifier.update.interval"));
      break;
  }
}

nsNotifierService.prototype.addObserver = function(aDomain, aFunction){
  var myPrefs = this.getPrefBranch();
  var prefBranchInternal = myPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);

  if (prefBranchInternal)
    prefBranchInternal.addObserver(aDomain, aFunction, false);
}

/* Bundle code */
nsNotifierService.prototype.getStringBundle = function() {
  if (!this.stringBundle) {
    var strBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"]
                             .createInstance(Components.interfaces.nsIStringBundleService);

    this.stringBundle = strBundleService.createBundle("chrome://gm-notifier/locale/gm-notifier.properties");
  }

  return this.stringBundle;
}

nsNotifierService.prototype.getString = function(aName) {
  return this.getStringBundle().GetStringFromName(aName);
}

nsNotifierService.prototype.getFormattedString = function(aName, aStrArray) {
  return this.getStringBundle().formatStringFromName(aName, aStrArray, aStrArray.length);
}

nsNotifierService.prototype.getLoginDetails = function(aUsername) {
  var url = "chrome://gm-notifier/";

  var passwordManager = Components.classes[kPSWDMANAGER_CONTRACTID]
                        .createInstance(nsIPasswordManagerInternal);
  var host = {value:""};
  var user =  {value:""};
  var password = {value:""};

  try {
    passwordManager.findPasswordEntry(url, aUsername, "", host, user, password);
  } catch(e){ }

  return password.value;
}

nsNotifierService.prototype.getURLProtocol = function() {
  var urlprotocol = "";
  if (this.getPrefBranch().getBoolPref("gm-notifier.connection.use.unsecured"))
    urlprotocol = "http://";
  else
    urlprotocol = "https://";

  return urlprotocol;
}



/**
 * JS XPCOM component registration goop:
 *
 * We set ourselves up to observe the xpcom-startup category.  This provides
 * us with a starting point.
 */

nsNotifierService.prototype.QueryInterface = function(iid) {
  if (!iid.equals(nsINotifierService) &&
      !iid.equals(Components.interfaces.nsIAlertListener) &&
      !iid.equals(Components.interfaces.nsIObserver) &&
      !iid.equals(nsISupports))
    throw Components.results.NS_ERROR_NO_INTERFACE;
  return this;
}

var gNotifierService = new nsNotifierService();

var nsNotifierServiceModule = new Object();

nsNotifierServiceModule.registerSelf = function (compMgr, fileSpec, location, type) {
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  compMgr.registerFactoryLocation(kGMSERVICE_CID,
                                  "nsIGMNotifierService",
                                  kGMSERVICE_CONTRACTID,
                                  fileSpec,
                                  location,
                                  type);
}

nsNotifierServiceModule.getClassObject = function (compMgr, cid, iid) {
  if (!cid.equals(kGMSERVICE_CID))
    throw Components.results.NS_ERROR_NO_INTERFACE;

  if (!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

  return nsNotifierServiceFactory;
}

nsNotifierServiceModule.canUnload = function (compMgr) {
  // cleanup
  gNotifierService.listeners = null;

  if (gNotifierService.updateTimer)
    gNotifierService.updateTimer.cancel();
  gNotifierService.updateTimer = null;

  return true;
}

var nsNotifierServiceFactory = new Object();

nsNotifierServiceFactory.createInstance = function (outer, iid) {
  if (outer != null)
    throw Components.results.NS_ERROR_NO_AGGREGATION;

  if (!iid.equals(nsINotifierService) &&
      !iid.equals(Components.interfaces.nsIAlertListener) &&
      !iid.equals(nsISupports))
    throw Components.results.NS_ERROR_NO_INTERFACE;

  return gNotifierService;
}

function NSGetModule(compMgr, fileSpec) {
  return nsNotifierServiceModule;
}

